package Zim::File;

use strict;
use Carp;
use File::Glob ':glob';
use File::Spec;
use File::Copy ();
use Encode;
use Zim::Events;

require Cwd;

our $VERSION = '0.18';
our @ISA = qw/Zim::Events/;

use overload
	'""' => \&path,
	fallback => 'TRUE' ;

unless (length $ENV{HOME}) {
	my ($home) = bsd_glob('~', GLOB_TILDE);
	$ENV{HOME} = $home if length $home;
}

=head1 NAME

Zim::File - OO wrapper for files

=head1 SYNOPSIS

	$file = Zim::File->new('./Changes');
	$text = $changes."\n\n".$file->read;
	$file->write($text);

=head1 DESCRIPTION

This package provides an object for handling files.
It bundles some snippets of code that are needed by several modules.

Inherits from L<Zim::Events>, supports a "write" event.

=head1 OVERLOAD

This class overloads the '""' operator, so the string version of an object
is the file path.

=head1 METHODS

=over 4

=item C<new(FILENAME)>

Simple constructor, uses L<File::Spec>'s C<catfile> when multiple arguments are given.

=cut

sub new {
	my $class = shift;
	my $name = (@_ == 1) ? shift : File::Spec->catfile(@_);
	croak "Invalid filename: '$name'" unless $name =~ /\S/;
	bless {path => $class->abs_path($name)}, $class;
}

=item C<set_check_mtime(BOOLEAN)>

Set whether to check mtime on read / write for this file.
When turned on an exception is raised when the mtime has
changed between read and write.

=cut

# feature turned off when undefined
# when "0" turned on but no mtime set yet

sub set_check_mtime {
	$_[0]{mtime} = $_[1] ? ($_[0]{mtime} || 0) : undef;
	$_[0]->signal_connect('write', \&_set_mtime) if $_[1];
}

=item C<path>

Returns the full path to the file.

=cut

sub path { $_[0]->{path} }

=item C<file>

Returns the filename part of the path.

=cut

sub file {
	my $self = shift;
	my $file = @_ ? File::Spec->rel2abs($_[0]) : $self->{path};
	my (undef, undef, $name) = File::Spec->splitpath($file);
	return $name;
}

=item C<dir>

Returns the directory part of the path.

=cut

sub dir {
	my $self = shift;
	my $file = @_ ? File::Spec->rel2abs($_[0]) : $self->{path};
	my ($vol, $dir, undef) = File::Spec->splitpath($file);
	return File::Spec->catpath($vol, $dir, '');
}

=item C<read()>

Returns the contents of the file. Takes list context in account.
Returns empty when the file does not exist.

=cut

sub read {
	my $self = shift;
	return unless $self->exists;
	my $fh = $self->open('r');
	my @text = <$fh>;
	$fh->close;
	return wantarray ? (@text) : join('', @text);
}

=item C<read_config()>

Read and parses contents.
Returns a hash with key value mappings and a list which indicates the order
in which the keys were found.

=cut

sub read_config {
	my $self = shift;
	my (%config, @order);
	for ($self->read) {
		/^\s*\#/ || ! /^\s*(.+?)=(.*)$/ and next;
		$config{$1} = $2;
		push @order, $1;
	}
	return wantarray ? (\%config, \@order) : \%config;
}

=item C<read_hash(KEY, CASE)>

Searches for KEY in a plain text hash.
The boolean CASE should be set for case sensitive lookup.
Case insensitive is the default.

=cut

sub read_hash {
	my ($self, $key, $case) = @_;
	$key = $case ? quotemeta($key) : qr/(?i:\Q$key\E)/ ;
	my $value = undef;
	my $fh = $self->open;
	while (<$fh>) {
		/^\s*$key/ or next;
		/^\s*$key\s+(.+?)\s*$/ or next;
		$value = $1;
		last;
	}
	$fh->close;
	return $value;
}

=item C<write(CONTENT, ..)>

Write CONTENT to file.
Creates file and parent directories when needed.

=item C<append(CONTENT, ..)>

Append content to file.
Creates file and parent directories when needed.

=cut

sub write {
	carp "## Zim::File::write: $_[0]";
	unshift @_, 'w';
	goto \&_write;
}

sub append {
	carp "## Zim::File::append: $_[0]";
	unshift @_, 'a';
	goto \&_write;
}

sub _write {
	my $mode = shift;
	my $self = shift;
	$self->make_dir if defined $self->{path};
	my $fh = $self->open($mode);
	print $fh @_;
	$fh->close or die "$!: $self->{path}\n";
}

=item C<write_config(\%CONFIG, \@ORDER)>

Writes a list of keys and values to file.
A second argument can be used to enforce the order of the keys in the
config file; this is optional.

=cut

sub write_config {
	my ($self, $config, $order) = @_;
	$order ||= [sort keys %$config];
	$self->write(map "$_=$$config{$_}\n", @$order);
}

=item C<remove()>

Delete the file.
Removes parent directories when empty.

=cut

sub remove {
	my $self = shift;
	my $file = $self->{path};
	unlink $file or die "Could not delete file: $file\n";
	$self->{mtime} = 0 if defined $self->{mtime};
	$self->remove_dir;
}

=item C<exists()>

Returns boolean.

=cut

sub exists { -f $_[0]->{path} }

=item C<stat()>

Returns a hash reference with stat values.

=cut

my @stat = qw/dev ino mode nlink uid gid rdev size
		atime mtime ctime blksize blocks/ ;

sub stat {
	my $self = shift;
	my @val = stat $self->{path};
	return { map {$stat[$_] => $val[$_]} 0 .. 12 };
}

=item C<open(MODE)>

Returns an C<IO> object that can be used as a file handle.

Make sure to close this handle using OO syntax, using C<< $fh->close >>
instead of C<close $fh>.

=cut

sub open {
	my ($self, $mode) = @_;
	$mode ||= 'r';
	#carp "## Zim::File::open($self, $mode)";
	my $file = $self->{path};
	if (defined $self->{mtime}) {
		my $mtime = (CORE::stat $file)[9];
		die "File has changed on disk since reading: $file\n"
			if $mode eq 'w'
			and $self->{mtime} and $self->{mtime} != $mtime ;
		$self->{mtime} = $mtime; 
	}
	my $handle = IO::File::Zim->new($file, $mode) || die "$!: $file\n";
	$handle->on_close(sub { $self->signal_emit('write', $file) })
		if $mode eq 'w';
	binmode $handle, ':utf8' unless $] < 5.008;
	return $handle;
}

sub _set_mtime { # called by 'write' event
	$_[0]{mtime} = (CORE::stat $_[0]{path})[9] if defined $_[0]{mtime};
}

=item C<grep(REGEXP, MODE)>

Line based grep function. MODE can be undefined, 'lines' or 'count';
In normal mode matches are returned, for 'lines' complete lines that
match are returns, and for 'count' a number of matches is returned.

=cut

sub grep {
	my ($self, $regexp, $mode) = @_;
	my $file = $self->{path};
	$mode ||= 'normal';
	return $mode eq 'count' ? 0 : () unless -f $file;
	#print STDERR "Grepping for /$regexp/ in $file (mode: $mode)\n\t";
	my @match;
	my $fh = $self->open;
	if ($mode eq 'lines') {
		while (<$fh>) {
			push @match, $_ if $_ =~ $regexp;
			#print STDERR '.';
		}
	}
	elsif ($mode eq 'count') {
		$match[0] = 0;
		while (<$fh>) {
			$match[0]++ if $_ =~ $regexp;
			#print STDERR '.';
		}
	}
	else {
		$regexp  = /($regexp)/;
		while (<$fh>) {
			push @match, ($_ =~ $regexp);
			#print STDERR '.';
		}
	}
	$fh->close;
	#print STDERR "DONE\n";
	#print STDERR "\t=> @match\n";
	return wantarray ? @match : $match[0] ;
}

=item C<touch()>

Create empty file if file does not exist.

=cut

sub touch {
	my $file = pop;
	carp "## Zim::File::touch: $file";
	return if -e $file;
	Zim::File->make_dir( Zim::File->dir($file) );
	CORE::open FH, '>', $file or die "$!: $file";
	print '';
	close FH;
}

=back

=head2 Class Methods

The following methods don't need an object to work on. They can also be called
as for example C<< Zim::File->make_dir($dir) >>. Of course they can also be
called on a Zim::File object.

=over 4

=item C<abs_path(NAME, REFERENCE)>

Turn a relative filename or a file uri into a local path name
and cleans up this path to a logical form. It also does tilde 
expansion.

If REFERENCE is omitted the current dir is used.

Calls C<parse_uri()> first when needed.

=cut

sub abs_path {
	my ($class, $file, $ref) = @_;
	#warn "abs_path: $name ($ref)\n";
	$file = $class->parse_uri($file) if $file =~ m#file:/#;
	return $file if $file =~ m#^\w\w+:/#;

	if ($file =~ m#^(~([^/]*))#) {
		my ($home) = length($2) ? bsd_glob($1, GLOB_TILDE) : $ENV{HOME};
		$file =~ s#^(~[^/]*)#$home#;
	}
	elsif ($file !~ m#^/#) {
		$ref ||= Cwd::cwd();
		$file = $ref.'/'.$file;
	}
	
	$file =~ s#(?<=[^/])/+\.(?![^/])##g; # remove /.
	while ($file =~ s#[^/]+/+\.\.(?![^/])##) {} # remove foo/..
	$file =~ s#^(/\.\.)+(?![^/])##; # remove leading /../
	$file =~ s#//+#/#g; # /// => /
	return $file;
}

=item C<rel_path(FILE, REFERENCE, UPWARD)>

Returns a path for FILE relative to REFERENCE
or undef when this is not possible.

When UPWARD is true the returned path can contain
"../", else it will only go below REFERENCE.

=cut

sub rel_path {
	my ($class, $file, $base, $upward) = @_;
	$file = $class->abs_path($file);
	$base = $class->abs_path($base);
	my @file = split '/', $file;
	my @base = split '/', $base;
	return undef if $base[0] ne $file[0];
		# might not even be same volume
	while (@base) {
		last unless $base[0] eq $file[0];
		shift @base;
		shift @file;
	}
	return undef if @base and ! $upward;
	if (@base) { unshift @file, '..' for @base }
	else       { unshift @file, '.'            }
	return join '/', @file;
}

=item C<parse_uri(URI)>

Method for parsing file uris, returns either a normal path or a (new) uri.
No url decoding is done here.

=cut

# support file:/path file://localhost/path and file:///path

sub parse_uri {
	shift; # class
	my $file = shift;
	
	$file =~ m#^file:/# or return $file;
	$file =~ s#^file:(?://localhost/+|/|///+)([^/]+)#/$1#i or return $file;

	return $file;
}

=item C<path2uri(FILE)>

Turns a path into an uri. Does no checks.

=cut

sub path2uri {
	shift; # class
	my $file = shift;
	$file =~ s#^/?#file:///#;
	return $file;
}

=item C<localize(FILE)>

Returns platform dependend version of FILE.

=cut

sub localize {return pop} # default does nothing, placeholder for win32

=item C<copy(SOURCE, DEST)>

Copy content from one file to another.
Creates parent directories if needed.

=cut

sub copy {
	my $self = $_[0];
	
	my $dest = pop;
	my $source = pop;
	$dest = $dest->path if ref $dest;
	$source = $source->path if ref $source;
	croak "Invalid file name: $dest\n" unless $dest =~ /\S/;
	croak "Invalid file name: $source\n" unless $source =~ /\S/;
	
	$self->make_dir($self->dir($dest));
	File::Copy::copy($source, $dest)
		or die "Could not copy '$source' to '$dest'\n";
}

=item C<move(SOURCE, DEST)>

Move file contents from one place to another.
Takes care of creating and cleaning up parent directories.

=cut

sub move {
	my $self = $_[0];
	
	my $dest = pop;
	my $source = pop;
	$dest = $dest->path if ref $dest;
	$source = $source->path if ref $source;
	croak "Invalid file name: $dest\n" unless $dest =~ /\S/;
	croak "Invalid file name: $source\n" unless $source =~ /\S/;
	
	$self->make_dir($self->dir($dest));
	#warn "moving $source => $dest\n";
	File::Copy::move($source, $dest)
		or die "Could not move '$source' to '$dest'\n";
	$self->remove_dir($self->dir($source));
}

=item C<list_dir(DIR)>

Returns a list of items for DIR. Ignores hidden files.

=cut

sub list_dir {
	my ($self, $dir) = @_;
	$dir = $self->dir if ! defined $dir and ref $self;
	croak "Invalid dir name: $dir" unless $dir =~ /\S/;
	
	opendir DIR, $dir or die "Could not list dir $dir\n";
	my @items = grep {! /^\./} readdir DIR;
	closedir DIR;

	eval {$_ = Encode::decode('utf8', $_, 1)} for @items;
	#warn "$dir =>\n", map "\t$_\n", @items;
	return @items;
}

=item C<grep_dir(DIR, REGEX)>

Returns all items in DIR that match REGEX.

=cut

sub grep_dir {
	my ($self, $dir, $regex) = @_;
	my @matches;
	opendir DIR, $dir or die "Could not open dir: $dir\n";
	while (my $item = readdir DIR) {
		eval { $item = Encode::decode('utf8', $item, 1) };
		push @matches, $item if $item =~ $regex;
	}
	closedir DIR;
	return @matches;
}

=item C<make_dir(DIR)>

Create DIR is it doesn't exist already. Subtle differences with L<File::Path>.

=cut

sub make_dir {
	my ($self, $dir) = @_;
	$dir = $self->dir if ! defined $dir and ref $self;
	croak "Invalid dir name: $dir" unless $dir =~ /\S/;
	
	return if -d $dir;
	
	my ($vol, $dirs) = File::Spec->splitpath(
		File::Spec->rel2abs($dir), 'DIR');
	my @dirs = File::Spec->splitdir($dirs);
	my $path = File::Spec->catpath($vol, shift @dirs, '');
	mkdir $path or die "Could not create dir: $path\n"
		unless ! length $path or -d $path;
	while (@dirs) {
		$path = File::Spec->catdir($path, shift @dirs);
		mkdir $path or die "Could not create dir $path\n"
			unless -d $path;
	}
}

=item C<remove_dir(DIR)>

Removes a directory and it's parents when they are empty.
Fails silently.

=cut

sub remove_dir {
	my ($self, $dir) = @_;
	$dir = $self->dir if ! defined $dir and ref $self;
	croak "Invalid dir name: $dir" unless $dir =~ /\S/;
	
	rmdir $dir or return; # fails when non-empty
	
	my ($vol, $dirs) = File::Spec->splitpath(
		File::Spec->rel2abs($dir), 'DIR');
	my @dirs = File::Spec->splitdir($dirs);
	while (@dirs) {
		pop @dirs;
		$dir = File::Spec->catdir($vol, @dirs);
		#warn "rmdir $dir\n";
		rmdir $dir or last; # fails when non-empty
	}
}

=item C<resolve_file(\%OPT, $ROOT, @PARTS)>

Look for a file below ROOT that kind of looks like @PARTS.
Options that can be passed in %OPT include:

=over 4

=item ext

File extension to add to the last part when resolving a file.

If a dir does not exist we try a file with this extension to resolve
the right case.

=item is_dir

We are looking for a dir.

=back

=cut

sub resolve_file {
	my ($class, $opt, $path, @parts) = @_;
	#warn "Resolve file: $path @parts\n";

	while (@parts) {
		my $part = shift @parts;
		my $match;
		if (-d $path) {
			my $is_file = (! @parts and ! $$opt{is_dir});
			$match = _grep_match($class, $path, $part, $$opt{ext}, $is_file);
		}
		if (defined $match) {
			#warn "\t$part => $match\n";
			$path .= '/' . $match;
		}
		else {
			$path = join '/', $path, $part, @parts;
			$path .= '.' . $$opt{ext} if ! $$opt{is_dir} and length $$opt{ext};
			last;
		}
	}

	#warn "\t=>$path\n";
	return $path;
}

sub _grep_match {
	my ($class, $path, $part, $ext, $is_file) = @_;
	my $regex = length($ext)
		? qr/^(?i:\Q$part\E)(?:$|\.$ext$)/
		: qr/^(?i:\Q$part\E)(?:$|\.)/  ;
	my @matches = $class->grep_dir($path, $regex);
	return undef unless @matches;

	if ($is_file and length $ext) { # prefer with extension
		my $f = $part . '.' . $ext;
		my @m = grep /\.\Q$ext\E$/, @matches;

		if (@m) {
			return $f if grep {$_ eq $f} @m;
			@m = sort @m;
			return $m[0];
		}
		else {
			return $f if grep {$_ eq $part} @matches;
			@matches = sort @matches;
			return $matches[0] . '.' . $ext;
		}
	}
	else {
		return $part if grep {$_ eq $part} @matches;

		@matches = sort @matches;
		return substr($matches[0], 0, length $part);
	}

	return undef;
}

package IO::File::Zim;

use IO::File;

our @ISA = qw/IO::File/;

# This class is a hack in order to provide a callback
# when the filehandle is closed. This is needed to fix
# stuff like saving the 'mtime' _after_ the file was writen.

# It uses a global hash to save object attributes because
# IO::Handle objects are filehandle references, not hashes.

# It uses a indirect subroutine definition to circumvent
# the Pod::Cover test

our %_objects;

*on_close = sub { # used to define a callback
	my $fh = shift;
	$_objects{"$fh"}{on_close} = shift;
};

*close = sub {
	my $fh = shift;
	my $re = $fh->SUPER::close(@_);
	return $re unless exists $_objects{"$fh"};
	my $obj = delete $_objects{"$fh"};
	$obj->{on_close}->() if $obj->{on_close};
	return $re;
};

1;

__END__

=back

=head1 AUTHOR

Jaap Karssenberg (Pardus) E<lt>pardus@cpan.orgE<gt>

Copyright (c) 2006 Jaap G Karssenberg. All rights reserved.
This program is free software; you can redistribute it and/or
modify it under the same terms as Perl itself.

=head1 SEE ALSO

L<Zim>

=cut

